/*
 *	emu.c
 *
 *	SUPERVISION/EMU
 *
 *	* Fri Apr 15 05:00:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

int
reg_read(int addr)
{
	switch(addr) {
	/*========== IRQ ==========*/
	case 0x2024:
		/* Timer IRQvNA܂B */
		ram[0x2027] &= ~(1<<0);
		return 0; /* ǂݏol͖Ӗł */
	case 0x2025:
		/* Sound DMA IRQvNA܂B */
		ram[0x2027] &= ~(1<<1);
		return 0; /* ǂݏol͖Ӗł */
	}
	return ram[addr];
}

void
reg_write(int addr, int data)
{
	ram[addr] = data;

	switch(addr) {
	/*========== Video ==========*/
	case 0x2000:
	case 0x2001:
	case 0x2002:
	case 0x2003:
		svvideo_write(addr, data);
		return;
	/*========== Sound ==========*/
	case 0x2010: /* Tone 0 */
	case 0x2011:
	case 0x2012:
	case 0x2013:
	case 0x2014: /* Tone 1 */
	case 0x2015:
	case 0x2016:
	case 0x2017:
	case 0x2018: /* DMA */
	case 0x2019:
	case 0x201a:
	case 0x201b:
	case 0x201c:
	case 0x2028: /* Noise */
	case 0x2029:
	case 0x202a:
		svsound_write(addr, data);
		return;
	/*========== I/O ==========*/
	case 0x2022:
		ram[0x2021] = data; /* $2022(W) -> $2021(R) */
		return;
	/*========== Timer ==========*/
	case 0x2023:
		if(!data) {
			data = 0x100; /* (YȂ!!) */
		}
		if(ram[0x2026] & (1<<4)) {
			TIMER = data << 15; /* x16384 clock */
		} else {
			TIMER = data << 8;  /* x  256 clock */
		}
		return;
	/*========== Bank,etc ==========*/
	case 0x2026:
		/* [0x2026] = bbbtldtn
		 *            |||||||+-- NMI enable
		 *            ||||||+--- Timer IRQ enable
		 *   +-++-----+++||+---- Sound DMA IRQ enable
		 *   | ||        |+----- LCD on (Ή)
		 *   | ||        +------ Timer mode
		 *   | ||
		 *   b bb000000 00000000 ... ROM bank address
		 */
		//ROM_BANK = &rom[(data & 0xe0) << 9] - 0x8000/*ROM window base address*/;
		//n[hEFAdl8bank=128KB܂ł̃J[gbW܂A
		//ۂɂ2bank=32KB4bank=64KB̃J[gbW݂܂B
		//ȒP̂߂ɁA[0x2026]̍ŏʃrbg0Œƍlď܂B
		ROM_BANK = &rom[(data & 0x60) << 9] - 0x8000/*ROM window base address*/;
		return;
	}
}

/****************************************************************************
 *	
 ****************************************************************************/

void
input_update()
{
	int pad = pcePadGetDirect();
	int input = 0xff;
	if(pad & PAD_RI    ) input &= ~(1<<0);
	if(pad & PAD_LF    ) input &= ~(1<<1);
	if(pad & PAD_DN    ) input &= ~(1<<2);
	if(pad & PAD_UP    ) input &= ~(1<<3);
	if(pad & PAD_B     ) input &= ~(1<<4);
	if(pad & PAD_A     ) input &= ~(1<<5);
	if(pad & PAD_SELECT) input &= ~(1<<6);
	if(pad & PAD_START ) input &= ~(1<<7);
	ram[0x2020/*input read*/] = input;
}

int
sv_init(const char* fname)
{
	FILE* fp;
	int size;

	/* ROMC[W[h܂B */
	fp = fopen(fname, "rb");
	if(!fp) {
		return -1;
	}
	size = fread(rom, 1, 0x10000, fp);
	fclose(fp);

	/* ROMC[WTCYɂ... */
	switch(size) {
	case  0x8000: /* 32KB */
		/* O32KB㔼32KBփ~[܂B */
		memcpy(&rom[0x8000], &rom[0x0000], 0x8000);
		break;
	case 0x10000: /* 64KB */
		/** no job **/
		break;
	default:
		return -1;
	}

	/* oNݒ肵܂B */
	ROM_BANK = &rom[0] - 0x8000/*ROM window base address*/;

	/* CA^C}NA܂B */
	memset(ram, 0, sizeof ram);
	TIMER = 0;

	/* Video/SoundZbg܂B */
	svvideo_reset(ram);
	svsound_reset(ram, rom);

	/* CPUZbg܂B */
	w65c02_reset();

	return 0;
}

int
sv_stream_callback(short* wbuff, int param)
{
	svsound_mix(wbuff);
	return 0;
}

int
sv_run(const char* fname, int sound_on)
{
	int cnt = 0;
	//
	int line;
	int irq;
	unsigned char* draw;

	if(sv_init(fname) != 0) {
		return -1;
	}

	if(sound_on) {
		stream_play(SVSOUNDBUFLEN, sv_stream_callback, 0, 0);
	}

	cnt = 0;
	for(;;) {
		/* ͍XVB */
		input_update();

		/* t[XLbvB */
		draw = NULL;
		if((cnt += FPS) >= 60) {
			cnt -= 60;
			draw = vbuff;
		}

		//-----------------------------------------------------------
		line = 160;
		do {
			w65c02_run(SUPERVISION_CLOCK/60/160);
			if(TIMER) {
				TIMER -= SUPERVISION_CLOCK/60/160;
				if(TIMER <= 0) {
					TIMER = 0;
					if(ram[0x2026] & (1<<1)) {	/* Timer IRQ enable? */
						ram[0x2027] |= (1<<0);	/* Timer IRQ */
					}
				}
			}
			/* IRQ𔻒肵܂B
			 * }XN(0x2026)Ɨv(0x2027)1bitYĂ邱ƂɒӂĂB
			 */
			irq = ram[0x2027/*D0=Timer,D1=AudioDMA*/] & 3;
			if(irq) { /* ̂߁A܂܂ɔ */
				irq &= ram[0x2026/*D1=Timer,D2=AudioDMA*/] >> 1;
				if(irq) {
					w65c02_irq();
				}
			}
		} while(--line);
		if(ram[0x2026] & 1) { /* NMI enable? */
			w65c02_nmi();
		}
		//-----------------------------------------------------------

		/* ʍXVB */
		if(draw) {
			svvideo_update(vbuff);
			svvideo_reduce(vbuff, vbuff);
			schedule();
			if((joy & PAD_START) && (joy & PAD_SELECT)) {
				break;
			}
		}
	}

	if(sound_on) {
		stream_stop();
	}

	return 0;
}

